/* liblxcapi
 *
 * SPDX-License-Identifier: LGPL-2.1+ *
 *
 * This function has been copied from musl.
 */

#ifndef _STRLCPY_H
#define _STRLCPY_H

#include "../lxc/compiler.h"
#include <stdio.h>

__hidden extern size_t strlcpy(char *, const char *, size_t);

#endif
